# pylint: disable=invalid-name

"""
Copyright (c) 2022 Troy Comi. All rights reserved.

abSENSE: Perform homolog detectability
"""
from __future__ import annotations

__version__ = "0.1.0"

__all__ = ("__version__",)
